### /u/sy/beebe/tex/bib/sutrs-to-bibtex.awk, Wed Nov 17 09:57:08 2004
### Edit by Nelson H. F. Beebe <beebe@math.utah.edu>
### ====================================================================
### Convert British Library SUTRS (Simple Unstructured Text Record
### Syntax) markup to rough BibTeX.
###
### Usage:
###	awk -f sutrs-to-bibtex.awk infile(s) > BibTeX-file
###
### [17-Nov-2004]
### ====================================================================

BEGIN				{ initialize() }

				{ gsub("[\r]+$", "") }

				{ gsub("\r", "") }

###				{ print "%% DEBUG:", $0 }

### Skip yaz-client chatter:

/Present request out of range/	{ next }
/^Connect/			{ next }
/^Diagnostic/			{ next }
/^Elapsed/			{ next }
/^ID /				{ next }
/^Name /			{ next }
/^Number of hits /		{ next }
/^Options/			{ next }
/^Received/			{ next }
/^Records/			{ next }
/^Search/			{ next }
/^Sent/				{ next }
/^Version/			{ next }
/^Z> /				{ next }
/^find /			{ next }
/^format /			{ next }
/^nextResultSetPosition/	{ next }
/^open /			{ next }
/^records returned/		{ next }
/^show/				{ next }
/^spawn/			{ next }

### Output previous data at start of new record

/\[BLAC\]Record/		{ print_entry(); next }

### Process the SUTRS markup that we understand

/^Acqns[.] Data/		{ ignore($0) ; next }
/^Add Rel\/Anal Title/		{ ignore($0) ; next }
/^Added Corporate/		{ ignore($0) ; next }
/^Added Person Name/		{ ignore($0) ; next }
/^Aut[.] Ctl[.] Info[.]/	{ ignore($0) ; next }
/^Authentication Code/		{ ignore($0) ; next }
/^Bibliography Note/		{ ignore($0) ; next }
/^Cataloging Source/		{ ignore($0) ; next }
/^DSC Acqns[.] Ind[.]/		{ ignore($0) ; next }
/^DSC Conf[.] Hdg./		{ do_conference(get_value($0)) ; next }
/^DSC Conf[.] Note/		{ do_remark(get_value($0)) ; next }
/^DSC Gen[.] Note:/		{ do_remark(get_value($0)) ; next }
/^DSC Shelvg[.] Ti[.]/		{ ignore($0) ; next }
/^Dewey DC Class[.] No[.]/	{ ignore($0) ; next }
/^Edition Statement/		{ do_edition(get_value($0)) ; next }
/^Edtg[.] Msg[.]/		{ ignore($0) ; next }
/^Elapsed/			{ ignore($0) ; next }
/^Finished Cataloguin/		{ ignore($0) ; next }
/^Foreign MARC/			{ ignore($0) ; next }
/^General Note/			{ ignore($0) ; next }
/^ID     /			{ ignore($0) ; next }
/^ISBN/				{ do_ISBN(get_value($0)) ; next }
/^Imprint/			{ do_imprint(get_value($0)) ; next }
/^Index Term- Unctrl/		{ do_subject(get_value($0)) ; next }
/^LC Control No[.]/		{ ignore($0) ; next }
/^LEO identifier/		{ ignore($0) ; next }
/^Legacy Ctl[.] No[.]/		{ ignore($0) ; next }
/^Link/				{ ignore($0) ; next }
/^Loc [&] Hol [(]temp[)]/	{ ignore($0) ; next }
/^Location/			{ ignore($0) ; next }
/^ME-Personal Name/		{ do_author(get_value($0)) ; next }
/^Name   /			{ ignore($0) ; next }
/^Nat[.]Bib[.]Iss[.]No[.]/	{ ignore($0) ; next }
/^National Bib[.] No[.]/	{ ignore($0) ; next }
/^Number of hits/		{ ignore($0) ; next }
/^Options/			{ ignore($0) ; next }
/^Phys[.]Description/		{ do_phys_description(get_value($0)) ; next }
/^Records/			{ ignore($0) ; next }
/^Series Statement/		{ do_series(get_value($0)) ; next }
/^Series-Title/			{ ignore($0) ; next }
/^Source/			{ ignore($0) ; next }
/^Stord[.]Srch[.]Flg[.]/	{ ignore($0) ; next }
/^Subject Cat[.] Code/		{ do_subject(get_value($0)) ; next }
/^Subject-Topical-LC/		{ do_subject(get_value($0)) ; next }
/^Subject-not specifi/		{ ignore($0) ; next }
/^System Control No[.]/		{ ignore($0) ; next }
/^Title/			{ do_title(get_value($0)) ; next }
/^Varying Form Title/		{ ignore($0) ; next }
/^Version/			{ ignore($0) ; next }

### Anything else gets logged as a comment for possible SUTRS pattern addition

				{ unrecognized($0) ; next }

END				{ terminate() }

### ====================================================================

function author_editor()
{
    if ("author" in Entry)
	return (Entry["author"])
    else if ("editor" in Entry)
	return (Entry["editor"])
    else
    {
	Entry["author"] = "Anonymous"
	return (Entry["author"])
    }
}


function capitalize(s, k,n,parts)
{
    if (is_corporate_name(s))
	return (s)		# leave corporate names alone
    if (s ~ /[A-Z][A-Z]/)	# Reduce HEARN, A. C. to Hearn, A. C.
    {				# but leave McMahon, L. E. unchanged
	n = split(s,parts," ")
	for (k = 1; k <= n; ++k)
	    s = (k == 1) ? \
		capitalize_word(parts[k]) : \
		s " " capitalize_word(parts[k])
    }
    return (s)
}


function capitalize_word(s, k)
{
    for (k = 1; k <= length(s); ++k)
    {
	# if (isupper(substr(s,k,1)) && (k > 1) && isletter(substr(s,k-1,1)))
	# [14-May-1996]: Change isletter() above to isupper(), so that
	# we do not convert "McLean" to "Mclean".
	if (isupper(substr(s,k,1)) && (k > 1) && isupper(substr(s,k-1,1)))
	    s = substr(s,1,k-1) tolower(substr(s,k,1)) substr(s,k+1)
    }
    return (s)
}


function citation_label( author,abbrev,k,n,parts,t,tag)
{
    author = author_editor()
    if (substr(author,1,1) == "{")
    {				# have corporate author
	split(author,parts," ")	# "{SPSS, Inc."}
	t = substr(parts[1],2)	# "SPSS,"
	gsub(/,/,"",t)		# reduce to "SPSS"
    }
    else
	t = capitalize(author)
    gsub(/ and /, ",", t)		# for uniform field separators
    split(t,parts,",")	# careful: "De La Beaujardiere, Jean-M" has spaces in name
    gsub(/[^-A-Za-z'\055]/,"",parts[1])	# strip non-(letters,apostrophe)

    sub(/Jr$/,"",parts[1])	# remove any Jr suffix
    tag = parts[1] ":" Entry["year"] ":"

    # We do a second strip here, because some entries for the year
    # can have spaces and other words in them, and because we do
    # not want apostrophes in the tag (bibxxx utilities complain).
    gsub(/[^-A-Za-z0-9:\055]/,"",tag)	# strip non-(letters,digits,colon))

    n = split(Entry["title"],parts," ")
    abbrev = ""
    for (k = 1; (k <= n) && (length(abbrev) < 3); ++k)
    {
	gsub(/[^A-Za-z0-9\'\055]/,"",parts[k]) # strip non-(letters,digits,apostrophe)
	if (isletter(substr(parts[k],1,1)))
	{
	    parts[k] = tolower(parts[k])
	    if (!(parts[k] in Ignore))
		abbrev = abbrev toupper(substr(parts[k],1,1))
	}
    }
    return (unique_citation_label(tag abbrev))
}


function clean_author_editor(s)
{
    # Typical values:
    # Smith, David, 1948 June 26-
    # Danuloff, Craig, 1963-
    # Dayton, Linnea

    gsub(/19[0-9][0-9]-? *([A-Z][a-z]+ [0-9]+-?)/,"",s)
    return (s)
}


function clean_entry()
{
    if ("author" in entry)
	set("author",clean_author_editor(Entry["author"]))

    if ("editor" in Entry)
	set("editor",clean_author_editor(Entry["editor"]))

    if ("title" in Entry)
    {
	gsub(/ *:/,":",Entry["title"]) # remove space before colons
	gsub(/ *\/[^\/]*$/,"",Entry["title"]) # strip author name(s) from title
	gsub("[ ]+[.][ ]+", " ", Entry["title"]) # strip isolated periods preceding subtitles
    }

    if ("edition" in Entry)
    {
	sub(/ ed[.]?$/,"",Entry["edition"])
	if (Entry["edition"] == "1st")
	    delete Entry["edition"]
	else if (Entry["edition"] == "2nd")
	    set("edition","Second")
	else if (Entry["edition"] == "3rd")
	    set("edition","Third")
	else if (Entry["edition"] == "4th")
	    set("edition","Fourth")
	else if (Entry["edition"] == "5th")
	    set("edition","Fifth")
	else if (Entry["edition"] == "6th")
	    set("edition","Sixth")
	else if (Entry["edition"] == "7th")
	    set("edition","Seventh")
	else if (Entry["edition"] == "8th")
	    set("edition","Eighth")
	else if (Entry["edition"] == "9th")
	    set("edition","Nineth")
	else if (Entry["edition"] == "10th")
	    set("edition","Tenth")
	else if (Entry["edition"] == "Rev.")
	    set("edition","Revised")
    }

    if ("ISBN" in Entry)
    {
	gsub(/ISBN */,"",Entry["ISBN"])
	clean_ISBN()
	Entry["ISBN"] = fix_ISBN(Entry["ISBN"])
    }

    if (("note" in Entry) && (substr(Entry["note"],length(Entry["note"]),1) != "."))
	Entry["note"] = Entry["note"] "."

    if ("price" in Entry)	# NB: must come after ISBN processing
    {
	gsub(/ *U[.]?S[.]?A[.]? */," ",Entry["price"])
	gsub(/ *,/,",",Entry["price"])
    }

    fix_annote()

    if ("author-dates" in Entry)
    {
	if (Entry["author-dates"] ~ "-- *$") # do not record dates for living authors
		delete Entry["author-dates"]
    }

    if ("note" in Entry)
	Entry["note"] = fix_note(Entry["note"])

    if ("pages" in Entry)
	Entry["pages"] = fix_pages(Entry["pages"])

    if ("LCCN" in Entry)
	Entry["LCCN"] = fix_LCCN(Entry["LCCN"])

    if ("subject" in Entry)
	Entry["subject"] = fix_subject(Entry["subject"])

    if ("URL" in Entry)
	Entry["URL"] = fix_URL(Entry["URL"])

    if ("year" in Entry)
	Entry["year"] = fix_year(Entry["year"])

    if (Bibsource != "")
	Entry["bibsource"] = Bibsource
}


function clean_ISBN( ISBN,parts)
{
    # Typical values
    #	"0070240493 (H) : US\$29.95, 0070240507 (P) : US\$19.95"
    #   "0442019963 (Van Nostrand Reinhold), 1850321418 (ITP UK)"
    #   "0125184085 (acid-free paper)"
    #   "0125531303 (alk. paper), 0125531311 (CD-Rom : alk. paper)"
    #   "0127378405 (alk. paper), 0127378413 (CD-ROM)"
    #   "0201633957 (alk. paper)"
    #   "0262112078 (alk. paper), 026261118X (pbk. : alk. paper)"
    #   "0471038571 (alk. paper : pbk.) : US\$19.95"
    #   "0838976972 (alk. paper) : US\$18.00 (LITA mbr. US\$16.20)"
    #   "0887369332 (alk. paper : pbk.) : US\$30.00 (US\$37.50 Can. : L21.00 UK)"
    #   "0471130885 (cloth : alk. paper)"
    #   "0841232237 (clothbound), 0841232245 (pbk.)"
    #   "0078811821 (cover), 0078821821: US\$34.95"
    #   "0079121004 (p)"
    #   "0471113360 (paper : acid-free paper) : US\$16.95"
    #   "0471128430 (paper : alk. paper)"
    #   "0712332952 (pbk)"
    #   "0471126942 (pbk. : acid-free paper)"
    #   "047114178X (pbk. : alk. paper)"
    #   "007024054X (pbk.) : US\$19.95"
    #   "020140978X (pbk.) US\$26.95 (US\$34.95 Can.)"
    #   "2707125024 (pbk.) : 140FF"
    #   "0192627058 (pbk.)"
    #   "0131515659 (recycled paper)"
    #   "0387945547 (softcover : alk. paper)"
    #   "0446671797 (trade pbk.)"
    #   "088736943X (v. 1 : alk. paper) : US\$7.00, 0887369448 (v. 2 : alk. paper) : US\$7.00, 0887369456 (v. 3 : alk. paper) : US\$7.00, 0887369464 (v. 4 : alk. paper) : US\$7.00, 0887369472 (v. 5 : alk. paper) : US\$7.00, 0887369480 (v. 6 : alk. paper) : US\$7.00"
    #   "012553132X, 0125531338 (CD-ROM)"
    #   "1556158858, 1556158181"
    #   "0070240558 : US\$16.95"
    #   "0070670188 : US\$45.00, 0070670196 (pbk.) : US\$29.95"
    #   "0077079051 : L22.95"
    #   "013185612X : US\$29.95 (US\$39.95 Can.)"
    #   "0201410001 : US\$17.95 (Can. US\$22.95)"
    #   "0201627078 : US\$14.95 (US\$18.95 Can.)"
    #   "0672305305 : US\$25.00 (US\$34.95 CAN)"
    #   "1562054376 : US\$35.00 (US\$47.95 Can.) (L32.49 UK)"
    #   "156761521X : US\$12.99"
    #   "0201407663 US\$12.95 (US\$16.95 Can.)"
    #   "0789701375 US\$39.99 (US\$53.99 Can.)"

    ISBN = entry["ISBN"]

    gsub(/[$]/,"US\\$",ISBN)

    gsub(/ +,/,",",ISBN)
    gsub(/\(v[.] /,"(vol. ",ISBN)
    gsub(/acid-free paper/,"",ISBN)
    gsub(/alk[.] paper/,"",ISBN)
    gsub(/mbr[.]/,"members",ISBN)
    gsub(/pbk[.]/,"paperback",ISBN)
    gsub(/pbk/,"paperback",ISBN)
    gsub(/\( *\)/,"",ISBN)
    gsub(/\([pP]\)/,"(paperback)",ISBN)
    gsub(/\([hH]\)/,"(hardcover)",ISBN)
    gsub(/ *: *\)/,")",ISBN)
    gsub(/\( *: */,"(",ISBN)
    gsub(/CD-Rom/,"CD-ROM",ISBN)
    gsub(/C[Aa][Nn][Aa][Dd][Aa] */,"Can.",ISBN)
    gsub(/C[Aa][Nn][.]? *US/,"CDN",ISBN)

    gsub(/[ :]+$/,"",ISBN)
    gsub(/[ ]+/," ",ISBN)

    while (match(ISBN,/US\\[$][0-9.]+ C[Aa][Nn][.]/))
    {
	ISBN = substr(ISBN,1,RSTART-1) "CDN" substr(ISBN,RSTART+2,RLENGTH-7) \
	  substr(ISBN,RSTART+RLENGTH)
    }

    while (match(ISBN,/US\\[$][0-9.]+ C[Aa][Nn]/))
    {
	ISBN = substr(ISBN,1,RSTART-1) "CDN" substr(ISBN,RSTART+2,RLENGTH-6) \
	  substr(ISBN,RSTART+RLENGTH)
    }

    while (match(ISBN,/L[0-9.]+/))
    {
	ISBN = substr(ISBN,1,RSTART-1) "UK\\pounds" substr(ISBN,RSTART+1,RLENGTH-1) \
	  substr(ISBN,RSTART+RLENGTH)
    }

    while (match(ISBN,/[0-9] : UK/))
    {
	ISBN = substr(ISBN,1,RSTART) "," substr(ISBN,RSTART+3)
    }

    while (match(ISBN,/\(CDN\\\$[0-9.]+\)/))
    {
	ISBN = substr(ISBN,1,RSTART-1) ", " substr(ISBN,RSTART+1,RLENGTH-2)
    }

    while (match(ISBN,/\(UK\\pounds[0-9.]+\)/))
    {
	ISBN = substr(ISBN,1,RSTART-1) ", " substr(ISBN,RSTART+1,RLENGTH-2)
    }

    gsub(/[ ]+,/,",",ISBN)
    gsub(/[ :]+$/,"",ISBN)
    gsub(/[ ]+/," ",ISBN)

    if ((index(ISBN,":") == 0) && \
	match(ISBN,/[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9xX],?( \([^)]*\))? *[A-Z]/))
	ISBN = substr(ISBN,1,RSTART + RLENGTH - 2) ": " substr(ISBN,RSTART + RLENGTH - 1)

    if (split(ISBN,parts,":") == 2)
    {
	ISBN = parts[1]

	gsub(/^ +/,"",parts[2])
	gsub(/[ ]+/," ",parts[2])
	entry["price"] = entry["price"] ", " parts[2]
	sub(/^, */,"",entry["price"])
    }

    gsub(/[ ,]+$/,"",ISBN)
    entry["ISBN"] = ISBN
}


function clear_array(array, key)
{
    for (key in array)
	delete array[key]
}


function delete_match(s)
{
    ## After a match() call on s, return s after deleting the match
    return (substr(s,1,RSTART-1) substr(s,RSTART+RLENGTH))
}


function do_author(s)
{
    ## Typical value:
    ##	Knuth, Donald E.   (Donald Ervin),   1938-

    if (match(s, " *[12][0-9][0-9][0-9]-$")) # drop birthdate for living authors
	s = delete_match(s)

    if ("author" in Entry)
	Entry["author"] = Entry["author"] " and " s
    else
	Entry["author"] = s
}


function do_conference(s)
{
    if ("conference" in Entry)
	Entry["conference"] = Entry["conference"] "; " s
    else
	Entry["conference"] = s
}


function do_edition(s)
{
    sub(" *[Ee]d[.]? *", "", s)
    if (s ~ "^1st")
	return
    else if (s == "2nd")
	s = "Second"
    else if (s == "3rd")
	s = "Third"
    else if (s == "4th")
	s = "Fourth"
    else if (s == "5th")
	s = "Fifth"
    else if (s == "6th")
	s = "Sixth"
    else if (s == "7th")
	s = "Seventh"
    else if (s == "8th")
	s = "Eighth"
    else if (s == "9th")
	s = "Nineth"
    else if (s == "10th")
	s = "Tenth"
    if ("edition" in Entry)
	Entry["edition"] = Entry["edition"] "; " s
    else
	Entry["edition"] = s
}


function do_imprint(s,   publisher,n,parts,year)
{
    ## Typical values:
    ##	London :   Robinson,   2003.
    ##	Reading :   X/Open Company Ltd,   [1993?]
    ##	Reading :   X/Open Company,   c1993.
    ##	Reading :   X/Open,   1992.
    ##	San Jose, California :   Adobe Press,   1997.
    ##	Silicon Press,   c1988.
    ##	Stanford, Calif. :   CSLI and Cambridge University Press,   c1996.
    ##	Upper Saddle River, N.J. :   Prentice Hall PTR,   c2004.

    n = split(s, parts, " *: *")
    if (n == 2)
    {
	s = parts[1]
	if ("address" in Entry)
	    Entry["address"] = Entry["address"] "; " s
	else
	    Entry["address"] = s

	s = parts[2]

	sub("[.]$", "", s)
        if (match(s, "c[12][0-9][0-9][0-9]$")) # "c1997"
	{
	    year = substr(s, RSTART + 1, RLENGTH - 1)
	    s = delete_match(s)
	}
        else if (match(s, "[12][0-9][0-9][0-9]$")) # "1997"
	{
	    year = substr(s, RSTART, RLENGTH)
	    s = delete_match(s)
	}
        else if (match(s, "\\[[12][0-9][0-9][0-9][?]*\\]$")) # [1997]"
	{
	    year = substr(s, RSTART + 1, RLENGTH - 2)
	    s = delete_match(s)
	}
	else
	    year = "????"

	year = substr(year, 1, 4)

	if ("year" in Entry)
	    Entry["year"] = Entry["year"] " / " year
	else
	    Entry["year"] = year

	publisher = trim(s)
	sub(" *,$", "", publisher)

	if ("publisher" in Entry)
	    Entry["publisher"] = Entry["publisher"] "; " publisher
	else
	    Entry["publisher"] = publisher
    }
    else				# save verbatim imprint information
    {
	if ("imprint" in Entry)
	    Entry["imprint"] = Entry["imprint"] "; " s
	else
	    Entry["imprint"] = s
    }
}


function do_ISBN(s,   price)
{
    ## Typical value:
    ##	1841196509 (pbk.) :   \XB98.99
    ##	1859120709 (pbk) :   No price
    ##	1881526925 (cased)
    ##	1881526917 (pbk)

    sub(" *: *No *price *$", "", s)

    if (match(s, "\\\\XB[0-9.]+$"))
    {
	price = substr(s, RSTART, RLENGTH)
	s = delete_match(s)
	sub("XB", "pounds ", price)
	if ("price" in Entry)
	    Entry["price"] = Entry["price"] "; " price
	else
	    Entry["price"] = price
    }

    if ("ISBN" in Entry)
	Entry["ISBN"] = Entry["ISBN"] ", " s
    else
	Entry["ISBN"] = s
}


function do_phys_description(s,   pages)
{
    ## Typical values
    ##	255 p. :   ill. ;   20 cm.
    ##	xii, 274p. :   ill., facsims. ;   24cm.
    ##	xvi,192p. ;   28cm.

    pages = s
    sub("[:;].*$", "", pages)
    sub(" *, *", " + ", pages)
    sub("p+[.]? *$", "", pages)

    pages = tolower(pages)

    if ("pages" in Entry)
	Entry["pages"] = Entry["pages"] "; " pages
    else
	Entry["pages"] = pages
}


function do_remark(s)
{
    if ("remark" in Entry)
	Entry["remark"] = Entry["remark"] ". " s
    else
	Entry["remark"] = s
}


function do_series(s,    series,volume)
{
    series = s

    if (match(series, " * *; *no[.] *[0-9A-Za-z]+ *$"))
    {
	volume = substr(series, RSTART, RLENGTH)
	series = delete_match(series)
	sub("^ * *; *no[.] *", "", volume)
	if ("volume" in Entry)
	    Entry["volume"] = Entry["volume"] "; " volume
	else
	    Entry["volume"] = volume
    }

    if ("series" in Entry)
	Entry["series"] = Entry["series"] "; " series
    else
	Entry["series"] = series
}


function do_subject(s)
{
    if ("subject" in Entry)
	Entry["subject"] = Entry["subject"] "; " s
    else
	Entry["subject"] = s
}


function do_title(s)
{
    sub("[.] *$", "", s)
    if ("title" in Entry)
	Entry["title"] = Entry["title"] "; " s
    else
	Entry["title"] = s
}


function enquote(v)
{
    if (substr(v,1,1) == "\"")	# enquote leading quote
	v = "{\377}" substr(v,2)
    while (match(v,/[^\\]"/))
    {				# brace unbackslashed quotes
	v = substr(v,1,RSTART) "{\377}" substr(v,RSTART+2)
    }
    gsub(/\377/,"\"",v)

    gsub(/&/," and ",v)		# replace ampersands
    gsub(/[ \t][ \t]+/," ",v)	# eliminate redundant spaces
    gsub(/[%]/,"\\%",v)		# protect quotes
    gsub(/[#]/,"\\#",v)		# protect sharps
    gsub(/[_]/,"\\_",v)		# protect underscores
    ## gsub(/[\^]/,"\\^",v)	# protect carets

    if (v ~ /[0-9][$][0-9]/)
	gsub(/[$]/,"--",v)	# June 7$10 -> June 7--10
#    else if (v ~ /[$][0-9]+/)
#	gsub(/[$]/,"US\\$",v)	# $27 -> US\$27 and $27.95 -> US\$27.95
    return (v)
}


function entry_name( )
{
    return (("entrytype" in Entry) ? \
	Entry["entrytype"] : "????UNKNOWN-TYPE-????")
}


function fix_annote( n,parts,e,s)
{
    # Remove some unwanted phrases in annote fields.

    if ("annote" in Entry)
    {
	s = Entry["annote"]
	gsub(/Includes bibliograph?ic(al)? references *(and index(es)?)? *(\([^\)]*\))?[.]?/,"",s)
	gsub(/Includes bibliograph?ic(al)? references *(p[.])? *(\([^\)]*\))? *(and index(es)?)?[.]?/,"",s)
	gsub(/Includes bibliograph?ic(al)? references *(and index(es)?)?[.]?/,"",s)
	gsub(/Includes bibliograph?(ies|y) *(and index(es)?)?[.]?/,"",s)
	gsub(/Includes glossar(y|ies) *(and index(es)?)?[.]?/,"",s)
	gsub(/Includes (author)? *index(es)?[.]?/,"",s)
	gsub(/Includes bibliograph?y[.]? *\([^\)]*\)[.]?/,"",s)
	gsub(/Includes bibliograph?(ic|ical|y) references[.]? *(\([^\)]*\))?(, *glossar(ies|y))? *(and *index(es)?)?[.]?/,"",s)
	gsub(/Includes bibliograph?(y|ies) \([^\)]*\) *(and index(es)?)[.]?*/,"",s)

	gsub(/Title from cover[.]?/,"",s)
	gsub(/Bibliograph?(y|ies) *: *(leaves|pages|pp[.]?|p[.]?) [0-9\-]*[.]?/,"",s)
	gsub(/Index *included[.]?/,"",s)
	gsub(/Index[.]?/,"",s)
	gsub(/ *and index[.]?/,"",s)

	if (!( ("month" in Entry) && ("day" in Entry) ))
	{
	    if (match(s,
		      /(January|February|March|April|May|June|July|August|September|October|November|December) *[123][0-9]?, *[12][0-9][0-9][0-9]/))
	    {
		n = split(substr(s,RSTART,RLENGTH),parts," ")
		if ((n == 3) && !("year" in Entry))
		    Entry["year"] = parts[3]
		if ((n == 3) && (parts[3] == Entry["year"]))
		{
		    sub(/,$/,"",parts[2])
		    Entry["day"] = parts[2]
		    Entry["month"] = month_abbrev[tolower(parts[1])]
		}
	    }
	    else if (match(s,
		      /[123][0-9]? *(January|February|March|April|May|June|July|August|September|October|November|December), *[12][0-9][0-9][0-9]/))
	    {
		n = split(substr(s,RSTART,RLENGTH),parts," ")
		if ((n == 3) && !("year" in Entry))
		    Entry["year"] = parts[3]
		if ((n == 3) && (parts[3] == Entry["year"]))
		{
		    sub(/,$/,"",parts[2])
		    Entry["day"] = parts[1]
		    Entry["month"] = month_abbrev[tolower(parts[2])]
		}
	    }
	}

	if (s ~ /^[ ]*$/)
	    delete Entry["annote"]
	else
	    set("annote",s)
    }
}


function fix_author_editor_names(s)
{
    ## Reduce "Henry. McGilton and Rachel. Morgan" to
    ## "Henry McGilton and Rachel Morgan"
    ## and
    ## "Haenel, Christoph, $d 1939-" to "Haenel, Christoph"

    ## print "%% DEBUG: enter fix_author_editor_names(\"" s "\")"

    if (match(s, "^ *and +"))
	s = substr(s, RSTART + RLENGTH)

    if (match(s, "[a-z][.]"))
	s = substr(s, 1, RSTART + RLENGTH - 2) "  " fix_author_editor_names(substr(s, RSTART + RLENGTH))

    s = (s " ")
    if (match(s, "[$]d +[12][0-9][0-9][0-9]-+[^0-9]"))
	s = substr(s, 1, RSTART - 1) fix_author_editor_names(substr(s, RSTART + RLENGTH - 1))

    ## print "%% DEBUG: leave fix_author_editor_names(): [" s "]"

    gsub("[ ]+[$][a-z][ ]+", " ", s)	# remove MARC subfield markers
    return (squeeze_blanks(s))
}


function fix_author_order()
{
    if ("author" in Entry)
	Entry["author"] = fix_names(Entry["author"])
}


function fix_editor_order()
{
    if ("editor" in Entry)
	Entry["editor"] = fix_names(Entry["editor"])
}


function fix_ISBN(s)
{
    # Typical values:
    ##	"0070450013 (pbk.)"
    ##	"0201134470 :, 0201134489 (pbk)",
    ##	"3540665722 (softcover : acid-free paper)"

    gsub(" *:? *acid-free *paper", "", s)
    gsub("[(] *pbk[.]? *[)]", "(paperback)", s)
    gsub("[(] *pbk[.]? *: *alk[.] *paper[)]", "(paperback)", s)
    gsub(" *: *, *", ", ", s)
    gsub("[(] *[)]", "", s)

    while (match(s, "[-0-9][-0-9][-0-9]+x"))
	s = (substr(s, 1, RSTART - 1) toupper(substr(s, RSTART, RLENGTH)) substr (s, RSTART + RLENGTH))

    return (s)
}


function fix_LCCN(s)
{
    # Typical value: "QA76.76.O63 $b M68 1987"
    sub(" +[$]b +", " ", s)
    return (s)
}


function fix_name_order(s, has_junior,has_others,has_sir,junior,n,parts,t)
{
    # Typical values (numbers indicate parts after splitting at commas)
    #
    # Davis, Mitchell, Jr		-> Mitchell {Davis, Jr.}	123 -> 2 {1, 3}
    # Murphy, Harry M. , Jr		-> Harry M. {Murphy, Jr.}	123 -> 2 {1, 3}
    # Donnelly, William J., II		-> William J. {Donnelly, II}	123 -> 2 {1, 3}
    # Reed, J. F., III			-> J. F. {Reed, III}		123 -> 2 {1, 3}
    #
    # Walsh, John E.			-> John E. Walsh		12 -> 2 1
    # Walsh, John E. (John Edward)	-> John E. (John Edward) Walsh	12 -> 2 1
    # Beethoven, Ludwig von		-> Ludwig von Beethoven		12 -> 2 1
    #
    # von Beethoven, Ludwig		-> Ludwig {von Beethoven}	12 -> 2 {1}
    # Auf Der Heide, Friedhelm Meyer	-> Friedhelm Meyer {Auf Der Heide} 12 -> 2 {1}
    #
    # This case requires special handling because the word order differs from the others:
    # Beerbohm, Max, Sir		-> Sir Max Beerbohm
    #
    # This one requires an extra brace, because part 1 contains a non-letter:
    # Butterworth IV, John F		-> John F. {Butterworth IV}	12 -> 2{1}
    #
    # This one requires special handling of Jr:
    # Lanahan, Robert J. Jr		-> Robert J. {Lanahan, Jr.}	12 -> xxxx
    # This case suggests that we always handle Jr specially.
    #
    # Notice that people like Guy L. Steele Jr. who omit the comma
    # before Jr must be coded in the library catalog like this:
    # Steele Jr., Guy L.
    # instead of
    # Steele, Guy L., Jr.
    # A search of the OCLC WorldCat database for this author showed that
    # in every case, the Jr is omitted entirely.

    gsub(/[ ,]*$/,"",s)		# remove trailing space and comma
    gsub(/[.] +[)]/,".)",s)	# "(Stuart I. )" -> "(Stuart I.)"

    t = trim(s)

    has_sir = match(t,/[, ]Sir$/) || match(t,/[, ]Count$/) || match(t,/[, ]Lord$/)
    has_junior = match(t,/[, ][JS]r[.]?$/) || match(t,/[, ][IVX]$/)

    if (has_junior)
    {
	junior = substr(t,RSTART+1,RLENGTH-1)
	if (junior ~ /^[JS][r]$/) # supply Jr./Sr. abbreviation period
	    junior = junior "."
	t = substr(t,1,RSTART-1) # remove Jr
	sub(/[, ]*$/, "", t)	# remove trailing space and comma (but NOT periods)
    }

    has_others = match(t,/[, ]et al[.]?/)
    if (has_others)
	t = delete_match(t)

    n = split(t, parts, /, */)

    parts[1] = trim(parts[1])
    if (match(parts[1]," "))
	parts[1] = "{" parts[1] "}"

    if (is_corporate_name(s))
	t = "{" trim(s)	"}"	# only brace corporate authors
    else if (n > 3)
    {
	warning("Cannot reorder a personal name with more than 3 parts: [" sample(s) "]")
	t = trim(s)
    }
    else if (n == 3)
    {
	if (has_junior)
	    t = parts[2] " {" parts[1] ", " parts[3] ", " junior "}"
	else if (has_sir)
	    t = parts[3] " " parts[2] " " parts[1]
	else
	    t = parts[2] " {" parts[1] ", " parts[3] "}"
    }
    else if (n == 2)
    {
	if (has_junior)
	    t = parts[2] " {" parts[1] ", " junior "}"
	else
	    t = parts[2] " " parts[1]
    }
    else
	t = s

    if (has_others)
	t = t " and others"
    return (t)
}


function fix_names(s, k,n,person,t)
{
    n = split(s, person, " and ")
    t = fix_name_order(person[1])
    for (k = 2; k <= n; ++k)
	t = t " and " fix_name_order(person[k])
    return (t)
}


function fix_note(s)
{
    # Typical value: "Includes index."

    ## print "%% DEBUG: enter fix_note(\"" s "\")"
    sub("Includes index[.] *", "", s)
    ## print "%% DEBUG: leave fix_note(): s = [" s "]"

    return (s)
}


function fix_pages(s)
{
    # Typical value: "xxv, 612 p. :"

    ## print "%% DEBUG: enter fix_pages(\"" s "\")"
    sub(" *: *$", "", s)
    sub(" +[0-9]+ *cm[.]? *$", "", s)
    sub(" *p[.]? *$", "", s)
    sub(" *p[.]? *ill[.]? *$", "", s)
    if (match(s, "^ *[ivxlcdm]+ *, *[0-9]+"))
	sub(" *, *", " + ", s)
    ## print "%% DEBUG: leave fix_pages(): s = [" s "]"

    return (s)
}


function fix_publisher()
{
    if (("publisher" in Entry) && (Entry["publisher"] == "Project Gutenberg"))
    {
	if (Entry_Count == 1)
	{
	    print "%-----------------------------------------------------------------------"
	    print "% Publisher abbreviations:"
	    print ""
	    print "@String{pub-PROJECT-GUTENBERG   = \"Project Gutenberg\"}"
	    print "@String{pub-PROJECT-GUTENBERG:adr = \"P.O. Box 2782,"
	    print "                                     Champaign, IL 61825-2782,"
	    print "                                     USA\"}"
	    print ""
	}
	Entry["publisher"] = "pub-PROJECT-GUTENBERG"
	if (!("address" in Entry))
	    Entry["address"] = "pub-PROJECT-GUTENBERG:adr"
    }
}


function fix_quotes(s, count,k)
{
    count = 0
    k = index(s, "\"")
    while (k > 0)
    {
	count++
	if ((count % 2) == 1)	# open quote
	    s = substr(s,1,k-1) "``" substr(s,k+1)
	else			# close quote
	    s = substr(s,1,k-1) "''" substr(s,k+1)
	k = index(s, "\"")
    }
    if ((count % 2) == 1)
	warning("Unbalanced double quotes in [" sample(s) "]")
    return (s)
}


function fix_series()
{
    if (("series" in Entry) && (Entry["series"] == "Project Gutenberg"))
    {
	if (Entry_Count == 1)
	{
	    print "%-----------------------------------------------------------------------"
	    print "% Series abbreviations:"
	    print ""
	    print "@String{ser-PROJECT-GUTENBERG = \"Project Gutenberg\"}"
	    print ""
	    print "%-----------------------------------------------------------------------"
	    print "% Bibliography entries:"
	    print ""
	}
	Entry["series"] = "ser-PROJECT-GUTENBERG"
    }
    gsub("[ ]+[$][a-z][ ]+", " ", Entry["series"])	# remove MARC subfield markers
}


function fix_subject(s)
{
    ## Typical value: "TeX (Computer file); Computerized typesetting.; Mathematics printing $x Computer programs."
    gsub(" +[$][a-z] +", "; ", s)
    gsub(" *[.];", ";", s)
    sub(" *[.] *$", "", s)
    return (s)
}


function fix_URL(s, j,k,n,words)
{

    ## The URLs are often duplicated, so eliminate the extra ones
    n = split(s, words, "; *")
    s = words[1]
    for (k = 2; k <= n; ++k)
    {
	dups = 0
	for (j = 1; j < k; ++j)
	{
	    if (words[j] == words[k])
	        break
	}
	if (j >= k)
	    s = s "; " words[k]
    }
    return (s)
}


function fix_year(s)
{
    ## print "%% DEBUG: enter fix_year(\"" s "\")"
    s = trim(s)
    sub("[.]+$", "", s)

    while (match(s, "\\[[12][0-9][0-9][0-9]\\]"))
	s = (substr(s, 1, RSTART - 1) substr(s, RSTART + 1, RLENGTH - 2) substr(s, RSTART + RLENGTH))

    sub("[.]+$", "", s)
    if (s ~ "^c[12][0-9][0-9][0-9]$")
	s = substr(s, 2,4)
    ## print "%% DEBUG: leave fix_year(): s = [" s "]"
    return (s)
}


function get_value(s)
{
    return (trim(substr(s, index(s,":") + 1)))
}


function ignore(s)
{
    print "%% IGNORED: ", s
}

function initialize()
{
    # NB: This MUST include a trailing slash
    URL_Base = "ftp://uiarchive.cso.uiuc.edu/pub/etext/gutenberg/"

    "date" | getline Current_Date_and_Time
    close("date")

    # For the Project Gutenberg USMARC list, which is ordered by
    # descending volume number, there are several labels which appear
    # more than once.  This table records their frequencies for use in
    # unique_citation_label() to generate a unique suffix letter.

    initialize_label_frequency()
    initialize_ignore_list()

    # Multiple USMARC records of the same number are joined with these
    # suffixes separating them:

    Suffix["abstract"] 		= ". "
    Suffix["author"] 		= " "
    Suffix["editor"] 		= " "
    Suffix["ISBN"] 		= ", "
    Suffix["ISSN"] 		= ", "
    Suffix["LCCN"] 		= "; "
    Suffix["language"] 		= "; "
    Suffix["note"] 		= ". "
    Suffix["publisher"] 	= " "
    Suffix["series"] 		= "; "
    Suffix["subject"] 		= "; "
    Suffix["title"] 		= " "
    Suffix["translator"] 	= "; "
    Suffix["URL"] 		= "; "
    Suffix["URL-contents"] 	= "; "
    Suffix["USMARC"] 		= "; "
    Suffix["USMARC-700"] 	= "; "
    Suffix["USMARC-740"] 	= "; "
    Suffix["USMARC-745"] 	= "; "
}


function initialize_ignore_list()
{
    # List of words to ignore in forming citation tags.  The initial
    # list was extracted from the bibindex badwords list, and covers
    # a few European languages as well as English.
    Ignore["a"]         = 1
    Ignore["ab"]        = 1
    Ignore["aber"]      = 1
    Ignore["als"]       = 1
    Ignore["an"]        = 1
    Ignore["and"]       = 1
    Ignore["are"]       = 1
    Ignore["as"]        = 1
    Ignore["auf"]       = 1
    Ignore["aus"]       = 1
    Ignore["az"]        = 1
    Ignore["bei"]       = 1
    Ignore["bir"]       = 1
    Ignore["but"]       = 1
    Ignore["da"]        = 1
    Ignore["das"]       = 1
    Ignore["dat"]       = 1
    Ignore["de"]        = 1
    Ignore["dei"]       = 1
    Ignore["dem"]       = 1
    Ignore["den"]       = 1
    Ignore["der"]       = 1
    Ignore["des"]       = 1
    Ignore["det"]       = 1
    Ignore["di"]        = 1
    Ignore["die"]       = 1
    Ignore["dos"]       = 1
    Ignore["e"]         = 1
    Ignore["een"]       = 1
    Ignore["eene"]      = 1
    Ignore["egy"]       = 1
    Ignore["ei"]        = 1
    Ignore["ein"]       = 1
    Ignore["eine"]      = 1
    Ignore["einen"]     = 1
    Ignore["einer"]     = 1
    Ignore["eines"]     = 1
    Ignore["eit"]       = 1
    Ignore["el"]        = 1
    Ignore["en"]        = 1
    Ignore["er"]        = 1
    Ignore["es"]        = 1
    Ignore["et"]        = 1
    Ignore["ett"]       = 1
    Ignore["eyn"]       = 1
    Ignore["eyne"]      = 1
    Ignore["for"]       = 1
    Ignore["from"]      = 1
    Ignore["fuer"]      = 1
    Ignore["fur"]       = 1
    Ignore["gl"]        = 1
    Ignore["gli"]       = 1
    Ignore["ha"]        = 1
    Ignore["haben"]     = 1
    Ignore["had"]       = 1
    Ignore["hai"]       = 1
    Ignore["has"]       = 1
    Ignore["hat"]       = 1
    Ignore["have"]      = 1
    Ignore["he"]        = 1
    Ignore["heis"]      = 1
    Ignore["hen"]       = 1
    Ignore["hena"]      = 1
    Ignore["henas"]     = 1
    Ignore["het"]       = 1
    Ignore["hin"]       = 1
    Ignore["hinar"]     = 1
    Ignore["hinir"]     = 1
    Ignore["hinn"]      = 1
    Ignore["hith"]      = 1
    Ignore["ho"]        = 1
    Ignore["hoi"]       = 1
    Ignore["i"]         = 1
    Ignore["il"]        = 1
    Ignore["in"]        = 1
    Ignore["is"]        = 1
    Ignore["ist"]       = 1
    Ignore["ka"]        = 1
    Ignore["ke"]        = 1
    Ignore["l"]         = 1
    Ignore["la"]        = 1
    Ignore["las"]       = 1
    Ignore["le"]        = 1
    Ignore["les"]       = 1
    Ignore["lo"]        = 1
    Ignore["los"]       = 1
    Ignore["mia"]       = 1
    Ignore["mit"]       = 1
    Ignore["n"]         = 1
    Ignore["na"]        = 1
    Ignore["nji"]       = 1
    Ignore["not"]       = 1
    Ignore["o"]         = 1
    Ignore["oder"]      = 1
    Ignore["of"]        = 1
    Ignore["on"]        = 1
    Ignore["or"]        = 1
    Ignore["os"]        = 1
    Ignore["others"]    = 1
    Ignore["s"]         = 1
    Ignore["sie"]       = 1
    Ignore["sind"]      = 1
    Ignore["so"]        = 1
    Ignore["t"]         = 1
    Ignore["ta"]        = 1
    Ignore["the"]       = 1
    Ignore["to"]        = 1
    Ignore["um"]        = 1
    Ignore["uma"]       = 1
    Ignore["un"]        = 1
    Ignore["una"]       = 1
    Ignore["und"]       = 1
    Ignore["une"]       = 1
    Ignore["uno"]       = 1
    Ignore["unter"]     = 1
    Ignore["von"]       = 1
    Ignore["with"]      = 1
    Ignore["y"]         = 1
    Ignore["yr"]        = 1

#   Additional words added later
    Ignore["also"]      = 1
    Ignore["any"]       = 1
    Ignore["away"]      = 1
    Ignore["by"]        = 1
    Ignore["cum"]       = 1
    Ignore["dans"]      = 1
    Ignore["down"]      = 1
    Ignore["into"]      = 1
    Ignore["its"]       = 1
    Ignore["off"]       = 1
    Ignore["onto"]      = 1
    Ignore["out"]       = 1
    Ignore["over"]      = 1
    Ignore["sur"]       = 1
    Ignore["that"]      = 1
    Ignore["these"]     = 1
    Ignore["this"]      = 1
    Ignore["those"]     = 1
    Ignore["unto"]      = 1
    Ignore["up"]        = 1
    Ignore["via"]       = 1
    Ignore["without"]   = 1
    Ignore["zu"]        = 1
    Ignore["zum"]       = 1
    Ignore["zur"]       = 1
}


function initialize_label_frequency()
{
    Label_Frequency["Anonymous:1995:USC"]	= 2
    Label_Frequency["Anonymous:1996:GWU"]	= 12
    Label_Frequency["Anonymous:1996:SRO"]	= 6
    Label_Frequency["Anonymous:1997:LVE"]	= 2
    Label_Frequency["Anonymous:1998:HBD"]	= 2
    Label_Frequency["Anonymous:1998:NMF"]	= 2
    Label_Frequency["Appleton:1997:TSH"]	= 6
    Label_Frequency["Appleton:1998:TSA"]	= 2
    Label_Frequency["Appleton:1998:TSH"]	= 6
    Label_Frequency["Ascham:1999:S"]		= 2
    Label_Frequency["Balzac:1998:A"]		= 2
    Label_Frequency["Balzac:1998:G"]		= 2
    Label_Frequency["Balzac:1998:MC"]		= 2
    Label_Frequency["Balzac:1998:P"]		= 2
    Label_Frequency["Balzac:1998:V"]		= 2
    Label_Frequency["Beethoven:1994:BFS"]	= 2
    Label_Frequency["Chesterton:1999:MWK"]	= 2
    Label_Frequency["Chesterton:1999:TP"]	= 2
    Label_Frequency["Conrad:1996:AF"]		= 2
    Label_Frequency["DanteAlighieri:1997:DCD"]	= 8
    Label_Frequency["DanteAlighieri:1997:HFC"]	= 4
    Label_Frequency["DanteAlighieri:1997:LTD"]	= 4
    Label_Frequency["Defoe:1996:RC"]		= 2
    Label_Frequency["Dickens:1998:MLL"]		= 2
    Label_Frequency["Gaskell:1999:LCB"]		= 2
    Label_Frequency["Gibbon:1997:HDF"]		= 2
    Label_Frequency["Gogol:1997:DS"]		= 2
    Label_Frequency["Homer:1999:O"]		= 2
    Label_Frequency["Lamb:1998:TS"]		= 2
    Label_Frequency["MacKay:1996:MPD"]		= 2
    Label_Frequency["Mansfield:1998:GP"]	= 2
    Label_Frequency["Marlowe:1997:DF"]		= 2
    Label_Frequency["Phillips:1996:C"]		= 2
    Label_Frequency["Plato:1998:C"]		= 2
    Label_Frequency["Plato:1998:L"]		= 2
    Label_Frequency["Plato:1999:A"]		= 2
    Label_Frequency["Plato:1999:C"]		= 2
    Label_Frequency["Plato:1999:E"]		= 2
    Label_Frequency["Plato:1999:M"]		= 2
    Label_Frequency["Plato:1999:P"]		= 5
    Label_Frequency["Plato:1999:S"]		= 3
    Label_Frequency["Prescott:1998:HCP"]	= 2
    Label_Frequency["Riley:1996:CWJ"]		= 2
    Label_Frequency["Rostand:1998:CB"]		= 3
    Label_Frequency["Shakespeare:1997:KHI"]	= 2
    Label_Frequency["Shakespeare:1997:KHV"]	= 5
    Label_Frequency["Shakespeare:1997:KRI"]	= 2
    Label_Frequency["Shakespeare:1997:SS"]	= 2
    Label_Frequency["Shakespeare:1998:C"]	= 2
    Label_Frequency["Shakespeare:1998:KHI"]	= 2
    Label_Frequency["Shakespeare:1998:KHV"]	= 5
    Label_Frequency["Shakespeare:1998:KRI"]	= 2
    Label_Frequency["Shakespeare:1998:MAA"]	= 2
    Label_Frequency["Shakespeare:1998:PP"]	= 2
    Label_Frequency["Shakespeare:1998:RL"]	= 2
    Label_Frequency["Shakespeare:1998:TN"]	= 2
    Label_Frequency["Shakespeare:1999:C"]	= 2
    Label_Frequency["Shakespeare:1999:KHI"]	= 2
    Label_Frequency["Shakespeare:1999:KHV"]	= 2
    Label_Frequency["Shakespeare:1999:KRI"]	= 2
    Label_Frequency["Shakespeare:1999:TA"]	= 2
    Label_Frequency["Spinoza:1997:E"]		= 5
    Label_Frequency["Stevenson:1992:DJM"]	= 2
    Label_Frequency["Stevenson:1996:LRL"]	= 2
    Label_Frequency["Stevenson:1996:SS"]	= 2
    Label_Frequency["Verne:1999:SC"]		= 2
    Label_Frequency["Virgil:1995:B"]		= 2
    Label_Frequency["Wharton:1995:ESF"]		= 2
    Label_Frequency["Wilde:1998:S"]		= 2
    Label_Frequency["Xenophon:1998:A"]		= 3
    Label_Frequency["Xenophon:1998:H"]		= 3
    Label_Frequency["Xenophon:1998:S"]		= 2
}


function is_corporate_name(s)
{
    return (match(s,/Ltd/) || match(s,/Co[.]/) || match(s,/Company/) || \
	    match(s,/Corp/) || match(s,/Inc[.]?$/) || match(s,/Pty[.]?/) || \
	    match(s,/Cie[.]/) || match(s,/AG/))
}


function isletter(c)
{
    return (index("ABCDEFGHIJKLMNOPQRSTUVWXYZ",toupper(c)) > 0)
}


function isupper(c)
{
    return (index("ABCDEFGHIJKLMNOPQRSTUVWXYZ",c) > 0)
}


function new_entry()
{
    clear_array(Entry)
    # Author_Editor_Key = "author"
    # Last_Key = ""
    # Entry["address"] = ""
}


function print_entry( k,key)
{
    if (!("title" in Entry))
    {
	new_entry()
	return
    }

    Entry_Count++
    k = 0
    for (key in Entry)		# count the number of saved key/value pairs
    {
	Entry[key] = fix_quotes(Entry[key])
	k++
    }

    set("entrytype","Book")	# NB: fixed type for now

    if ("author" in Entry)
	Entry["author"] = fix_author_editor_names(Entry["author"])

    if ("editor" in Entry)
	Entry["editor"] = fix_author_editor_names(Entry["editor"])

    if (("entrytype" in Entry) && (k > 0))
    {
	# These two handle the special case of Project Gutenberg USMARC records
	# and must be called before any output is done
	fix_publisher()
	fix_series()

	clean_entry()

	if (entry_name() == "InProceedings")
	    print_Proceedings()	# want InProceedings to crossref a Proceedings

	print_line("@" entry_name() "{" citation_label() ",")

	if (entry_name() == "Phone")
	{
	    delete Entry["author"]
	    delete Entry["title"]
	}

	# These two must come AFTER the call to citation_label(), because
	# that function assumes the original "Last, First Middle" order
	# of names, while these functions change them to BibNet standard
	# order "First Middle Last".

	fix_author_order()
	fix_editor_order()

	if (("publisher" in Entry) && (Entry["publisher"] == "pub-PROJECT-GUTENBERG"))
	    Entry["acknowledgement"] = "ack-nhfb # \" and \" # ack-st"
	else
	    Entry["acknowledgement"] = "ack-nhfb"
	Entry["bibdate"] = Current_Date_and_Time

	# Change "not available" to something more recognizable
	if (("volume" in Entry) && (Entry["volume"] == "na"))
	    Entry["volume"] = "n/a"

	# After trying strict alphabetic order for the keywords, I
	# decided that it is more useful to have most of them in a
	# specific order, that corresponding to the usual order in an
	# article citation.  print_pair() will delete each entry that
	# it prints, and ignore any that don't exist, so that key/value
	# pairs are not duplicated in the subsequent output after
	# sort_keys().  The key order here is chosen to match biborder(1).

	print_pair("author")
	print_pair("editor")
	print_pair("key")
	print_pair("booktitle")
	print_pair("title")
	print_pair("crossref")
	print_pair("chapter")
	print_pair("journal")
	print_pair("volume")
	print_pair("type")
	print_pair("number")
	print_pair("howpublished")
	print_pair("institution")
	print_pair("organization")
	print_pair("publisher")
	print_pair("school")
	print_pair("address")
	print_pair("edition")
	print_pair("pages_whole")	# NB: Must precede "pages" because it
	print_pair("pages")		# checks for the existence of "pages"
	print_pair("day")
	print_pair("month")
	print_pair("year")
	print_pair("ISBN")
	print_pair("ISSN")
	print_pair("LCCN")
	print_pair("bibdate")
	print_pair("bibsource")
	print_pair("note")
	print_pair("price")
	print_pair("series")
	print_pair("acknowledgement")
	print_pair("subject")

	# Output any remaining key/value pairs sorted by key name
	sort_keys()

	for (k = 1; Sorted_Keys[k]; ++k)
	    print_key_value(Sorted_Keys[k],Entry[Sorted_Keys[k]])

	print_line("}\n")
    }

    new_entry()
}


function print_key_abbrev(key,abbrev)
{
    print_line(sprintf("  %s =%s %s,", key, \
	substr("                ",1,12-length(key)), abbrev))
}


function print_key_string(key,string)
{
    # TeXify 8-bit accented characters
    gsub(/\252/,"\\?",string)	# what is this character?? Perhaps trademark
    gsub(/\260/,"\\?",string)	# what is this character?? Displays as Angstrom accent in willow, but as separate character
    gsub(/\341/,"\\`",string)
    gsub(/\342/,"\\'",string)
    gsub(/\343/,"\\^",string)
    gsub(/\346/,"\\?",string)	# what is this character??
    gsub(/\351/,"\\v ",string)
    gsub(/\365x/,"${}^x$",string)
    gsub(/\365X/,"${}^X$",string)
    gsub(/\350a/,"{\\\"a}",string)
    gsub(/\350e/,"{\\\"e}",string)
    gsub(/\350o/,"{\\\"o}",string)
    gsub(/\350u/,"{\\\"u}",string)

    gsub(/[.][.][.]/,"\\ldots{}",string) # convert ellipses ... to \ldots{}

    gsub(/[;:, ]+$/,"",string)	# remove selected trailing punctuation

    print_line(sprintf("  %s = %s\"%s\",", \
	key, \
	substr("                ",1,12-length(key)), \
	trim(enquote(string))))
}


function print_key_value(key,value)
{
    if (key == "entrytype")		# already output as @type{tag, # }
	return
    else if (key == "acknowledgement")
	print_key_abbrev(key,value)
    else if ((key == "publisher") && ("publisher" in Entry) && (Entry["publisher"] ~ "pub-"))
	print_key_abbrev(key,value)
    else if ((key == "series") && ("series" in Entry) && (Entry["series"] ~ "ser-"))
	print_key_abbrev(key,value)
    else if ((key == "address") && ("address" in Entry) && (Entry["address"] ~ "pub-"))
	print_key_abbrev(key,value)
    else if (key == "month")
	print_month(value)
    else if (key == "pages_whole")
	print_key_string(("pages" in Entry) ? "pageswhole" : "pages", value)
    else
	print_key_string(key,value)
}


function print_line(line)
{
    print line
}


function print_month(month, mon)
{
    warning("print_month not yet implemented")
    return

    mon = tolower(month)
    gsub(/[^A-Za-z\055]/,"",mon)	# strip punctuation
    if (mon in month_abbrev)
	print_key_abbrev("month", month_abbrev[mon])
    else
	print_key_string("month",month)
}


function print_pair(key)
{
    if (key in Entry)
    {
	if (Entry[key] != "")
	    print_key_value(key,Entry[key])
	delete Entry[key]
    }
}


function print_Proceedings()
{
    warning("print_Proceedings() not yet implemented")
}


function sample(s)
{
    # return up to 256 chars of s
    return ((length(s) > 256) ? (substr(s,1,256) "...") : s)
}


function set(key,value)
{
    Entry[key] = value
    Last_Key = key
}


function sort_keys( k,key,m,n)
{
    clear_array(Sorted_Keys)

    n = 0
    for (key in Entry)
    {
	n++
	Sorted_Keys[n] = key
    }
    for (k = 1; k < n; ++k)
    {
	for (m = k + 1; m <= n; ++m)
	{
	    if (tolower(Sorted_Keys[k]) > tolower(Sorted_Keys[m]))
	    {
		key = Sorted_Keys[m]
		Sorted_Keys[m] = Sorted_Keys[k]
		Sorted_Keys[k] = key
	    }
	}
    }
}


function squeeze_blanks(s)
{
    gsub(/[ \t][ \t]+/," ",s)
    return (s)
}


function terminate()
{
    if (("author" in Entry) || ("title" in Entry))
	print_entry();
}


function trim(s)
{
    gsub(/^[ \t]+/,"",s)
    gsub(/[ \t]+$/,"",s)
    return (s)
}


function unique_citation_label(label)
{
    # For the Project Gutenberg USMARC list, supply a suffix letter to
    # make citation labels unique.
    if ((label in Label_Frequency) && \
	("publisher" in Entry) && \
	(Entry["publisher"] == "pub-PROJECT-GUTENBERG"))
    {
	if (label in Label_Count)
	    Label_Count[label]--
	else
	    Label_Count[label] = Label_Frequency[label]

	if (Label_Count[label] <= 0)
	    warning("Label_Frequency[\"" label "\"] is wrong in initialize()")
	else
	    label = label substr("abcdefghijklmnopqrstuvwxyz", Label_Count[label], 1)
    }

    # Check for duplicate citation labels
    if (label in Label_Used)
	warning("duplicate citation label \"" label "\": add or increase Label_Frequency[] entry in initialize_label_frequency()")
    Label_Used[label]++

    return (label)
}


function unrecognized(s)
{
    print "%% UNRECOGNIZED SUTRS MARKUP: ", s
}


function warning(message)
{
    print FILENAME ":" FNR ":%%" message >"/dev/stderr"
}
